%{
/*-
 * Copyright (c) 2008, Florent Thoumie <flz@FreeBSD.org>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $FreeBSD$
 */

#include <sys/types.h>
#include <stdlib.h>
#include <stdio.h>
#include <err.h>

#include "lib.h"

enum { VAR_STRING, VAR_LIST };

struct _str_list_t {
	char **strs;
	int len;
};

int	yylex(void);
int	yyparse(void);
int	yyerror(const char *);

%}

%union {
	struct _str_list_t *list;
	char *str;
	int i;
}

%token <i> NAME
%token <str> STRING
%type <list> list
%token EQ QMARK COMMA OBRACE CBRACE
%token <str> PACKAGESITE PACKAGEROOT

%%

config	
	: 
	| options
       	;

options
	:
	| options option
	;

option
	: STRING EQ STRING {
		config_set_str($1, $3);
		free($1);
	}
	| STRING EQ OBRACE list CBRACE {
		struct _str_list_t *list = $4;
		config_set_list($1, list->strs);
		free($1);
	}
	;

list
	: STRING {
		struct _str_list_t *list = calloc(1, sizeof(struct _str_list_t));
		list->len = 2;
		list->strs = calloc(list->len, sizeof(char *));
		list->strs[0] = $1;
		$$ = list;
	}
	| list COMMA STRING {
		struct _str_list_t *list = $1;
		list->len++;
		list->strs = realloc(list->strs, list->len * sizeof(char *));
		list->strs[list->len-2] = $3;
		$$ = list;
	}
	;

%%

char **
config_get_list(const char *name)
{
	struct config_var *cv;

	SLIST_FOREACH(cv, &config, next)
		if (strncmp(cv->name, name, strlen(name)) == 0) {
			if (cv->type != VAR_LIST) {
				printf("config: %s should be of type 'list'. \
				    Exiting.", name);
				exit(1);
			} else {
				return cv->list;
			}
		}

	return NULL;
}

char *
config_get_str(const char *name)
{
	struct config_var *cv;

	SLIST_FOREACH(cv, &config, next)
		if (strncmp(cv->name, name, strlen(name)) == 0) {
			if (cv->type != VAR_STRING) {
				printf("config: %s should be of type 'string'. \
				    Exiting", name);
				exit(1);
			} else {
				return cv->str;
			}
		}

	return NULL;
}

void
config_set_list(char *name, char **list)
{
	struct config_var *cv;
	char *tmp;
	int i;

	if (list == NULL)
		return;

	tmp = strdup(name);
	str_lowercase(tmp);

	/* Find and replace ... */
	SLIST_FOREACH(cv, &config, next)
		if (strncmp(cv->name, tmp, strlen(tmp)) == 0) {
			if (cv->type == VAR_STRING) {
				free(cv->str);
			} else {
				for (i = 0; cv->list[i]; ++i)
					free(cv->list[i]);
				free(cv->list);
			}
			cv->type = VAR_LIST;
			cv->list = list;
			free(tmp);
			return ;
		}

	/* ... or add. */
	cv = calloc(1, sizeof(struct config_var));
	cv->name = tmp;
	cv->type = VAR_LIST;
	cv->list = list;
	SLIST_INSERT_HEAD(&config, cv, next);
}

void
config_set_str(char *name, char *str)
{
	struct config_var *cv;
	char *tmp;
	int i;

	if (str == NULL)
		return;

	tmp = strdup(name);
	str_lowercase(tmp);

	/* Find and replace ... */
	SLIST_FOREACH(cv, &config, next)
		if (strncmp(cv->name, tmp, strlen(tmp)) == 0) {
			if (cv->type == VAR_STRING) {
				free(cv->str);
			} else {
				for (i = 0; cv->list[i]; ++i)
					free(cv->list[i]);
				free(cv->list);
			}
			cv->type = VAR_STRING;
			cv->str = str;
			free(tmp);
			return ;
		}

	/* ... or add. */
	cv = calloc(1, sizeof(struct config_var));
	cv->name = tmp;
	cv->type = VAR_STRING;
	cv->str = str;
	SLIST_INSERT_HEAD(&config, cv, next);
}

static char **
str2list(char *str)
{
	char **list;

	if (str == NULL)
		return NULL;

	list = calloc(2, sizeof(char *));
	list[0] = str;
	return list;
}

extern FILE* yyin;
extern char configfile[FILENAME_MAX];

void
config_init(void)
{
    struct config_var *cv;
    char userconf[FILENAME_MAX];
    int i, error;

    SLIST_INIT(&config);

    config_set_str("dbdir", strdup("/var/db/pkg"));
    config_set_str("portsdir", strdup("/usr/ports"));
    config_set_str("packagesdir", strdup("/var/packages:/usr/ports/packages"));

    config_set_list("packagesite", str2list(getenv("PACKAGESITE")));
    config_set_list("packageroot", str2list(getenv("PACKAGEROOT")));
    config_set_str("tmpdir", getenv("TMPDIR"));
    config_set_str("tmpdir", getenv("PKG_TMPDIR"));
    config_set_str("dbdir", getenv("PKG_DBDIR"));
    config_set_str("portsdir", getenv("PORTSDIR"));
    config_set_str("packagesdir", getenv("PKGDIR"));

    yyin = fopen("/etc/pkg.conf", "r");
    if (yyin != NULL) {
        strncpy(configfile, "/etc/pkg.conf", FILENAME_MAX);
	error = yyparse();
	fclose(yyin);
    }

    yyin = fopen("/etc/pkg.conf.local", "r");
    if (yyin != NULL) {
        strncpy(configfile, "/etc/pkg.conf.local", FILENAME_MAX);
	error = yyparse();
	fclose(yyin);
    }

    snprintf(userconf, FILENAME_MAX, "%s/.pkgrc", getenv("HOME"));

    yyin = fopen(userconf, "r");
    if (yyin != NULL) {
        strncpy(configfile, userconf, FILENAME_MAX);
	error = yyparse();
	fclose(yyin);
    }

    if (Verbose) {
        SLIST_FOREACH(cv, &config, next) {
		if (cv->type == VAR_STRING)
			printf("config: %s = %s\n", cv->name, cv->str);
		else {
			printf("config: %s = { ", cv->name);
			for (i = 0; cv->list[i]; ++i)
			    printf("%s, ", cv->list[i]);
			printf(" }\n");
		}
	}
    }
}
